use palette::{
    cam16::{Cam16Jch, Parameters},
    convert::FromColorUnclamped,
    luma::Luma,
    Hsl, IntoColor, IsWithinBounds, Oklab, Oklch, Srgb,
};

pub fn map_cam16jch_to_srgb(origin: &Cam16Jch<f32>) -> Srgb {
    let mut new_original = Cam16Jch::new(origin.lightness, origin.chroma, origin.hue);
    const FACTOR: f32 = 0.99;
    loop {
        let new_srgb =
            Srgb::from_color_unclamped(new_original.into_xyz(Parameters::default_static_wp(40.0)));
        if new_srgb.is_within_bounds() {
            break new_srgb;
        }
        new_original = Cam16Jch::new(
            new_original.lightness,
            new_original.chroma * FACTOR,
            new_original.hue,
        );
    }
}

pub fn map_cam16jch_to_srgb_hex(origin: &Cam16Jch<f32>) -> String {
    format!("{:x}", map_cam16jch_to_srgb(origin).into_format::<u8>())
}

pub fn map_hsl_to_srgb(origin: &Hsl) -> Srgb {
    let mut new_original = Hsl::new(origin.hue, origin.saturation, origin.lightness);
    const FACTOR: f32 = 0.99;
    loop {
        let new_srgb = Srgb::from_color_unclamped(new_original);
        if new_srgb.is_within_bounds() {
            break new_srgb;
        }
        new_original = Hsl::new(
            new_original.hue,
            new_original.saturation * FACTOR,
            new_original.lightness,
        );
    }
}

pub fn map_hsl_to_srgb_hex(origin: &Hsl) -> String {
    format!("{:x}", map_hsl_to_srgb(origin).into_format::<u8>())
}

pub fn map_oklch_to_luma(origin: &Oklch) -> Luma {
    let lab_color: Oklab = (*origin).into_color();
    let linear_rgb = Srgb::from_linear(lab_color.into_color());
    let luma_color: Luma = linear_rgb.into_color();
    luma_color
}

pub fn map_oklch_to_srgb(origin: &Oklch) -> Srgb {
    Srgb::from_linear::<f32>((*origin).into_color())
}

pub fn map_oklch_to_srgb_hex(origin: &Oklch) -> String {
    format!("{:x}", map_oklch_to_srgb(origin).into_format::<u8>())
}

#[macro_export]
macro_rules! parse_to_oklch {
    ($origin: ident) => {
        palette::Oklch::from_color(
            palette::Srgb::from_str($origin)
                .map_err(|_| crate::errors::ColorError::UnrecogniazedRGB($origin.to_string()))?
                .into_format::<f32>(),
        )
    };
    ($origin: expr) => {
        palette::Oklch::from_color(
            palette::Srgb::from_str($origin)
                .map_err(|_| crate::errors::ColorError::UnrecogniazedRGB($origin.to_string()))?
                .into_format::<f32>(),
        )
    };
}

#[macro_export]
macro_rules! parse_option_to_oklch {
    ($origin: ident) => {
        $origin
            .map(|color| {
                let rgb = palette::Srgb::from_str(color)
                    .map_err(|_| crate::errors::ColorError::UnrecogniazedRGB(color.to_string()))?
                    .into_format::<f32>();
                Ok(palette::Oklch::from_color(rgb))
            })
            .transpose()?
    };
}
