use std::{
    ops::{Add, Sub},
    str::FromStr,
};

use palette::{color_difference::Wcag21RelativeContrast, FromColor, Hsl, Srgb};

use crate::{convert::map_hsl_to_srgb, errors};

#[derive(Debug, Clone)]
pub struct M2Swatch {
    _key_color: Hsl,
    swatch: Vec<Hsl>,
    desaturated_swatch: Vec<Hsl>,
    pub key_index: SwatchIndex,
}

#[derive(Debug, Clone, Copy)]
pub enum SwatchIndex {
    SI900,
    SI800,
    SI700,
    SI600,
    SI500,
    SI400,
    SI300,
    SI200,
    SI100,
    SI50,
}

impl SwatchIndex {
    pub fn to_usize(&self) -> usize {
        match self {
            SwatchIndex::SI900 => 0,
            SwatchIndex::SI800 => 1,
            SwatchIndex::SI700 => 2,
            SwatchIndex::SI600 => 3,
            SwatchIndex::SI500 => 4,
            SwatchIndex::SI400 => 5,
            SwatchIndex::SI300 => 6,
            SwatchIndex::SI200 => 7,
            SwatchIndex::SI100 => 8,
            SwatchIndex::SI50 => 9,
        }
    }

    pub fn from_usize(index: usize) -> Self {
        match index {
            0 => SwatchIndex::SI900,
            1 => SwatchIndex::SI800,
            2 => SwatchIndex::SI700,
            3 => SwatchIndex::SI600,
            4 => SwatchIndex::SI500,
            5 => SwatchIndex::SI400,
            6 => SwatchIndex::SI300,
            7 => SwatchIndex::SI200,
            8 => SwatchIndex::SI100,
            9 => SwatchIndex::SI50,
            _ => panic!("Invalid index"),
        }
    }
}

impl Add<i16> for SwatchIndex {
    type Output = Self;

    fn add(self, rhs: i16) -> Self::Output {
        let index = (self.to_usize() as i16 + rhs).clamp(0, 9);
        SwatchIndex::from_usize(index as usize)
    }
}

impl Add<i16> for &SwatchIndex {
    type Output = SwatchIndex;

    fn add(self, rhs: i16) -> Self::Output {
        let index = (self.to_usize() as i16 + rhs).clamp(0, 9);
        SwatchIndex::from_usize(index as usize)
    }
}

impl Sub<i16> for SwatchIndex {
    type Output = Self;

    fn sub(self, rhs: i16) -> Self::Output {
        let index = (self.to_usize() as i16 - rhs).clamp(0, 9);
        SwatchIndex::from_usize(index as usize)
    }
}

impl Sub<i16> for &SwatchIndex {
    type Output = SwatchIndex;

    fn sub(self, rhs: i16) -> Self::Output {
        let index = (self.to_usize() as i16 - rhs).clamp(0, 9);
        SwatchIndex::from_usize(index as usize)
    }
}

fn find_key_color_place(value: f32, min_value: f32, max_value: f32) -> usize {
    let interval = (max_value - min_value) / 9.0;
    let values = (0..=9)
        .map(|i| min_value + interval * i as f32)
        .collect::<Vec<f32>>();

    let mut closest_index = 0;
    let mut min_diff = (value - values[closest_index]).abs();

    for (i, v) in values.iter().enumerate() {
        let diff = (value - v).abs();
        if diff < min_diff {
            min_diff = diff;
            closest_index = i;
        }
    }

    closest_index
}

impl M2Swatch {
    pub fn default_neutral() -> Self {
        let key_color = Hsl::new(0.0, 0.0, 0.0);
        let mut swatch = Vec::new();

        let interval: f32 = 1.0 / 9.0;
        for i in 0..=9 {
            swatch.push(Hsl::new(0.0, 0.0, interval * i as f32));
        }

        Self {
            _key_color: key_color,
            desaturated_swatch: swatch.clone(),
            swatch,
            key_index: SwatchIndex::SI900,
        }
    }

    pub fn from_color(color: &str) -> Result<Self, errors::ColorError> {
        let key_color = Hsl::from_color(
            Srgb::from_str(color)
                .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
                .into_format::<f32>(),
        );
        let mut swatch = Vec::new();
        let mut desaturated_swatch = Vec::new();

        let max_lightness = key_color.lightness.max(0.95);
        let min_lightness = key_color.lightness.min(0.2);
        let max_staturation = key_color.saturation.max(0.80);
        let min_staturation = key_color.saturation.min(0.45);

        let key_place = find_key_color_place(key_color.lightness, min_lightness, max_lightness);

        if key_place > 0 {
            let lightness_interval = (key_color.lightness - min_lightness) / key_place as f32;
            let saturation_interval = (max_staturation - key_color.saturation) / key_place as f32;
            for i in 0..key_place {
                swatch.push(Hsl::new(
                    key_color.hue,
                    max_staturation - saturation_interval * i as f32,
                    min_lightness + lightness_interval * i as f32,
                ));
                desaturated_swatch.push(Hsl::new(
                    key_color.hue,
                    (min_staturation + saturation_interval * i as f32) * 0.7_f32,
                    min_lightness + lightness_interval * i as f32,
                ));
            }
        }

        swatch.push(key_color.clone());

        if key_place < 9 {
            let lightness_interval = (max_lightness - key_color.lightness) / (9 - key_place) as f32;
            let saturation_interval =
                (key_color.saturation - min_staturation) / (9 - key_place) as f32;
            for i in 1..=9 - key_place {
                swatch.push(Hsl::new(
                    key_color.hue,
                    key_color.saturation - saturation_interval * i as f32,
                    key_color.lightness + lightness_interval * i as f32,
                ));
                desaturated_swatch.push(Hsl::new(
                    key_color.hue,
                    (min_staturation + saturation_interval * i as f32) * 0.7_f32,
                    key_color.lightness + lightness_interval * i as f32,
                ))
            }
        }

        Ok(Self {
            _key_color: key_color,
            swatch,
            desaturated_swatch,
            key_index: SwatchIndex::from_usize(key_place),
        })
    }

    pub fn tone(&self, index: SwatchIndex) -> Hsl {
        self.swatch[index.to_usize()].clone()
    }

    pub fn key_tone(&self) -> Hsl {
        self.swatch[self.key_index.to_usize()].clone()
    }

    pub fn desaturated_tone(&self, index: SwatchIndex) -> Hsl {
        self.desaturated_swatch[index.to_usize()].clone()
    }

    pub fn desaturated_key_tone(&self) -> Hsl {
        self.desaturated_swatch[self.key_index.to_usize()].clone()
    }

    pub fn wacg_max_tone(&self, reference_color: &str) -> Result<Hsl, errors::ColorError> {
        let reference_color = Srgb::from_str(reference_color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(reference_color.to_string()))?
            .into_format::<f32>();

        let mut max_wacg_index = 0;
        let mut max_wacg = 0.0;

        for (i, tone) in self.swatch.iter().enumerate() {
            let swatch_color = map_hsl_to_srgb(tone);
            let wacg = swatch_color.relative_contrast(reference_color);
            if wacg > max_wacg {
                max_wacg = wacg;
                max_wacg_index = i;
            }
        }

        Ok(self.swatch[max_wacg_index].clone())
    }

    pub fn desaturated_wacg_max_tone(
        &self,
        reference_color: &str,
    ) -> Result<Hsl, errors::ColorError> {
        let reference_color = Srgb::from_str(reference_color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(reference_color.to_string()))?
            .into_format::<f32>();

        let mut max_wacg_index = 0;
        let mut max_wacg = 0.0;

        for (i, tone) in self.desaturated_swatch.iter().enumerate() {
            let swatch_color = map_hsl_to_srgb(tone);
            let wacg = swatch_color.relative_contrast(reference_color);
            if wacg > max_wacg {
                max_wacg = wacg;
                max_wacg_index = i;
            }
        }

        Ok(self.desaturated_swatch[max_wacg_index].clone())
    }

    pub fn wacg_min_above(
        &self,
        reference_color: &str,
        ratio: f32,
    ) -> Result<Hsl, errors::ColorError> {
        let reference_color = Srgb::from_str(reference_color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(reference_color.to_string()))?
            .into_format::<f32>();

        let mut min_wacg_index = 0;
        let mut min_wacg = 21.0;

        for (i, tone) in self.swatch.iter().enumerate() {
            let swatch_color = map_hsl_to_srgb(tone);
            let wacg = swatch_color.relative_contrast(reference_color);
            if wacg < min_wacg && wacg > ratio {
                min_wacg = wacg;
                min_wacg_index = i;
            }
        }

        Ok(self.swatch[min_wacg_index].clone())
    }

    pub fn desaturated_wacg_min_above(
        &self,
        reference_color: &str,
        ratio: f32,
    ) -> Result<Hsl, errors::ColorError> {
        let reference_color = Srgb::from_str(reference_color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(reference_color.to_string()))?
            .into_format::<f32>();

        let mut min_wacg_index = 0;
        let mut min_wacg = 21.0;

        for (i, tone) in self.desaturated_swatch.iter().enumerate() {
            let swatch_color = map_hsl_to_srgb(tone);
            let wacg = swatch_color.relative_contrast(reference_color);
            if wacg < min_wacg && wacg > ratio {
                min_wacg = wacg;
                min_wacg_index = i;
            }
        }

        Ok(self.desaturated_swatch[min_wacg_index].clone())
    }
}
